### Santalucia_NN_Tm; a part of Thermo-Align tool for the design of template specific hybridization and priming oligonucleotides
### Updated: version-1.05 06/29/2017
### Property of Wisser Lab at University of Delaware
### Author: Felix Francis (felixfrancier@gmail.com)
#    Copyright (C) 2016-2030 by
#    Felix Francis <felixfrancier@gmail.com>
#    All rights reserved.

############################################################
#### IMPORT FUNCTIONS
############################################################
import math

############################################################
#### FUNCTIONS AND TABLES
############################################################

# SantaLucia & Hicks (2004), Annu. Rev. Biophys. Biomol. Struct 33: 415-440
# delta H (Enthalpy)(kcal/mol) and delta S (Entropy)(eu) coefficients
DNA_NN_table = { 
	'init': (0.2, -5.7), 'init_A/T': (2.2, 6.9), 'init_G/C': (0, 0), 
	'init_oneG/C': (0, 0), 'init_allA/T': (0, 0), 'init_5T/A': (0, 0),
	'sym': (0, -1.4),
	'AA/TT': (-7.6, -21.3), 'AT/TA': (-7.2, -20.4), 'TA/AT': (-7.2, -20.4), 
	'CA/GT': (-8.5, -22.7), 'GT/CA': (-8.4, -22.4), 'CT/GA': (-7.8, -21.0), 
	'GA/CT': (-8.2, -22.2), 'CG/GC': (-10.6, -27.2), 'GC/CG': (-9.8, -24.4), 
	'GG/CC': (-8.0, -19.0)}

# Internal mismatch and inosine table (DNA) 
# Allawi & SantaLucia (1997), Biochemistry 36: 10581-10594 
# Allawi & SantaLucia (1998), Biochemistry 37: 9435-9444 
# Allawi & SantaLucia (1998), Biochemistry 37: 2170-2179 
# Allawi & SantaLucia (1998), Nucl Acids Res 26: 2694-2701 
# Peyret et al. (1999), Biochemistry 38: 3468-3477 					

DNA_IMM_table = { 
	'AG/TT': (1.0, 0.9), 'AT/TG': (-2.5, -8.3), 'CG/GT': (-4.1, -11.7), 				# Allawi & SantaLucia (1997), Biochemistry 36: 10581-10594  http://pubs.acs.org/doi/pdf/10.1021/bi962590c
	'CT/GG': (-2.8, -8.0), 'GG/CT': (3.3, 10.4), 'GG/TT': (5.8, 16.3), 
	'GT/CG': (-4.4, -12.3), 'GT/TG': (4.1, 9.5), 'TG/AT': (-0.1, -1.7), 
	'TG/GT': (-1.4, -6.2), 'TT/AG': (-1.3, -5.3),
		
	'AA/TG': (-0.6, -2.3), 
	'AG/TA': (-0.7, -2.3), 'CA/GG': (-0.7, -2.3), 'CG/GA': (-4.0, -13.2), 
	'GA/CG': (-0.6, -1.0), 'GG/CA': (0.5, 3.2), 'TA/AG': (0.7, 0.7), 
	'TG/AA': (3.0, 7.4),
		
	'AC/TT': (0.7, 0.2), 'AT/TC': (-1.2, -6.2), 'CC/GT': (-0.8, -4.5), 					# Allawi & SantaLucia (1998), Nucl Acids Res 26: 2694-2701 
	'CT/GC': (-1.5, -6.1), 'GC/CT': (2.3, 5.4), 'GT/CC': (5.2, 13.5), 
	'TC/AT': (1.2, 0.7), 'TT/AC': (1.0, 0.7),
		
	'AA/TC': (2.3, 4.6), 'AC/TA': (5.3, 14.6), 'CA/GC': (1.9, 3.7), 					# Allawi & SantaLucia (1998), Biochemistry 37: 9435-9444  http://pubs.acs.org/doi/pdf/10.1021/bi9803729 
	'CC/GA': (0.6, -0.6), 'GA/CC': (5.2, 14.2), 'GC/CA': (-0.7, -3.8), 
	'TA/AC': (3.4, 8.0), 'TC/AA': (7.6, 20.2),
		
	'AA/TA': (1.2, 1.7), 'CA/GA': (-0.9, -4.2), 'GA/CA': (-2.9, -9.8), 					# Peyret et al. (1999), Biochemistry 38: 3468-3477 
	'TA/AA': (4.7, 12.9), 'AC/TC': (0.0, -4.4), 'CC/GC': (-1.5, -7.2), 
	'GC/CC': (3.6, 8.9), 'TC/AC': (6.1, 16.4), 'AG/TG': (-3.1, -9.5), 
	'CG/GG': (-4.9, -15.3), 'GG/CG': (-6.0, -15.8), 'TG/AG': (1.6, 3.6), 
	'AT/TT': (-2.7, -10.8), 'CT/GT': (-5.0, -15.8), 'GT/CT': (-2.2, -8.4), 
	'TT/AT': (0.2, -1.5)} 

# Terminal mismatch table (DNA) 
# SantaLucia & Peyret (2001) Patent Application WO 01/94611 
DNA_TMM_table = { 
	'AA/TA': (-3.1, -7.8),	'TA/AA': (-2.5, -6.3),
	'CA/GA': (-4.3, -10.7),	'GA/CA': (-8.0, -22.5), 
	'AC/TC': (-0.1, 0.5),	'TC/AC': (-0.7, -1.3),
	'CC/GC': (-2.1, -5.1),	'GC/CC': (-3.9, -10.6), 
	'AG/TG': (-1.1, -2.1),	'TG/AG': (-1.1, -2.7),
	'CG/GG': (-3.8, -9.5),	'GG/CG': (-0.7, -19.2), 
	'AT/TT': (-2.4, -6.5), 'TT/AT': (-3.2, -8.9),
	'CT/GT': (-6.1, -16.9),	'GT/CT': (-7.4, -21.2), 
	'AA/TC': (-1.6, -4.0), 'AC/TA': (-1.8, -3.8), 'CA/GC': (-2.6, -5.9), 
	'CC/GA': (-2.7, -6.0), 'GA/CC': (-5.0, -13.8), 'GC/CA': (-3.2, -7.1), 
	'TA/AC': (-2.3, -5.9), 'TC/AA': (-2.7, -7.0), 
	'AC/TT': (-0.9, -1.7), 'AT/TC': (-2.3, -6.3), 'CC/GT': (-3.2, -8.0), 
	'CT/GC': (-3.9, -10.6), 'GC/CT': (-4.9, -13.5), 'GT/CC': (-3.0, -7.8), 
	'TC/AT': (-2.5, -6.3), 'TT/AC': (-0.7, -1.2), 
	'AA/TG': (-1.9, -4.4), 'AG/TA': (-2.5, -5.9), 'CA/GG': (-3.9, -9.6), 
	'CG/GA': (-6.0, -15.5), 'GA/CG': (-4.3, -11.1), 'GG/CA': (-4.6, -11.4), 
	'TA/AG': (-2.0, -4.7), 'TG/AA': (-2.4, -5.8), 
	'AG/TT': (-3.2, -8.7), 'AT/TG': (-3.5, -9.4), 'CG/GT': (-3.8, -9.0), 
	'CT/GG': (-6.6, -18.7), 'GG/CT': (-5.7, -15.9), 'GT/CG': (-5.9, -16.1), 
	'TG/AT': (-3.9, -10.5), 'TT/AG': (-3.6, -9.8)} 

# Dangling ends table (DNA) 
# Bommarito et al. (2000), Nucl Acids Res 28: 1929-1934 
DNA_DE_table = { 
	'AA/.T': (0.2, 2.3), 'AC/.G': (-6.3, -17.1), 'AG/.C': (-3.7, -10.0), 
	'AT/.A': (-2.9, -7.6), 'CA/.T': (0.6, 3.3), 'CC/.G': (-4.4, -12.6), 
	'CG/.C': (-4.0, -11.9), 'CT/.A': (-4.1, -13.0), 'GA/.T': (-1.1, -1.6), 
	'GC/.G': (-5.1, -14.0), 'GG/.C': (-3.9, -10.9), 'GT/.A': (-4.2, -15.0), 
	'TA/.T': (-6.9, -20.0), 'TC/.G': (-4.0, -10.9), 'TG/.C': (-4.9, -13.8), 
	'TT/.A': (-0.2, -0.5), 
	'.A/AT': (-0.7, -0.8), '.C/AG': (-2.1, -3.9), '.G/AC': (-5.9, -16.5), 
	'.T/AA': (-0.5, -1.1), '.A/CT': (4.4, 14.9), '.C/CG': (-0.2, -0.1), 
	'.G/CC': (-2.6, -7.4), '.T/CA': (4.7, 14.2), '.A/GT': (-1.6, -3.6), 
	'.C/GG': (-3.9, -11.2), '.G/GC': (-3.2, -10.4), '.T/GA': (-4.1, -13.1), 
	'.A/TT': (2.9, 10.4), '.C/TG': (-4.4, -13.1), '.G/TC': (-5.2, -15.0), 
	'.T/TA': (-3.8, -12.6)} 

### DELTA S ION CORRECTION FUNCTION
def ion_correction(Na=0, K=0, Tris=0, Mg=0, dNTPs=0, seq_len=None):
	'''
	Correction for deltaS(Entropy correction) : 0.368 x (N-1) x ln[Na+]
	Reference: (SantaLucia (1998), Proc Natl Acad Sci USA 95: 1460-1465)	http://www.pnas.org/content/95/4/1460.full.pdf+html
	Provide Millimolar(mmol/L) concentrations for Na, K, Tris, Mg and dNTPs
	Von Ahsen et al. (2001, Clin Chem 47: 1956-1961)	http://www.clinchem.org/content/47/11/1956.full
	[Na_eq] = [Na+] + [K+] + [Tris]/2 + 120*([Mg2+] - [dNTPs])^0.5 
	If [dNTPs] >= [Mg2+]: [Na_eq] = [Na+] + [K+] + [Tris]/2
	effect on entropy by salt correction; von Ahsen et al 1999
		s+=0.368 * (strlen(c)-1)* log(salt_effect)
	'''
	correction_factor		=	0
	Na=float(Na)
	K=float(K)
	Tris=float(Tris)
	Mg=float(Mg)
	dNTPs=float(dNTPs)
	Monovalent_cations_mmol	=	Na + K + (Tris/2)
	if dNTPs >= Mg:
		Na_eq_mmol		=	Monovalent_cations_mmol 
	else:
		Na_eq_mmol		=	Monovalent_cations_mmol + (120 * math.sqrt(Mg-dNTPs))
	Na_eq_mol	=	Na_eq_mmol/1000.0												# Convert from millimolar to molar
	correction_factor		=	0.368 * (seq_len - 1) * math.log(Na_eq_mol)
	return correction_factor

### FUNCTION TO CHECK WHETHER TWO SEQUENCES ARE REVERSE COMPLEMENTS OF EACH OTHER
def transform_degenerate(seq):
	seq = seq.upper()
	basesub = {'R': 'A',
		'Y': 'C',
		'M': 'A',
		'K': 'G',
		'S': 'G',
		'W': 'A',
		'H': 'A',
		'B': 'G',
		'V': 'G',
		'D': 'G'}
	for base in basesub.keys():
		seq = seq.replace(base, basesub[base])
	return seq

def rev_complement(seq):
	seq = seq.upper()
	basecomplement = {'A':'T', 
					  'C':'G', 
					  'G':'C', 
					  'T':'A', 
					  '-':'-', 
					  'N':'N'}
	letters = list(seq)
	newletters = []
	for base in letters:
		if base in basecomplement:
			newletters.append(basecomplement[base])
		else:
			newletters.append('N')
	complement = (''.join(newletters))                                              #gives the complement of the bases in list letters
	return complement[::-1]                                                         #gives the reverse of the compliment   

def complement(seq):
	seq = seq.upper()
	basecomplement = {'A':'T', 
					  'C':'G', 
					  'G':'C', 
					  'T':'A', 
					  '-':'-', 
					  'N':'N'}
	letters = list(seq)
	newletters = []
	for base in letters:
		if base in basecomplement:
			newletters.append(basecomplement[base])
		else:
			newletters.append('N')
	complement = (''.join(newletters))                                                 #gives the complement of the bases in list letters
	return complement 

### FUNCTION TO QC INPUT OLIGO SEQUENCES. (covert the input seq to uppercase, remove any whitespace and characters other than 'A','C','G','T','N','-')
def seq_qc(seq):
	seq = ''.join(seq.split()).upper()
	allowed_chars =('A','C','G','T','N','-')													# Reconsider including other allowed characters '-"???
	seq = ''.join([base for base in seq if base in allowed_chars])
	return seq

### FUNCTION TO CALCULATE millimolar (mM) CONCENTRATION OF MONOVALENT AND DIVALENT CATIONS (OUTPUT IS IN mM)
def mM_monovalent(Na=0, K=0, Tris=0, Mg=0, dNTPs=0):
	Na=float(Na)
	K=float(K)
	Tris=float(Tris)
	Mg=float(Mg)
	dNTPs=float(dNTPs)
	Monovalent_cations_mmol	=	Na + K + (Tris/2)
	if dNTPs >= Mg:
		Na_eq_mmol		=	Monovalent_cations_mmol 
	else:
		Na_eq_mmol		=	Monovalent_cations_mmol + (120 * math.sqrt(Mg-dNTPs))
	return Na_eq_mmol

### FUNCTION TO CALCULATE NN TERMODYNAMICS BASED MELTING TEMPERATURE(TM)

def NN_Tm(seq=None, compl_seq=None, primer_conc=100, Na=0, K=50, Tris=10, Mg=1.5, dNTPs=0.2, ion_corr=True):
	""" Calculates the Tm using the nearest neighbor thermodynamics
	Arguments:
		-seq		: The primer sequence 			(5'->3' direction)
		-compl_seq	: The complementary sequence	(3'->5' direction!!!!!) 
		-primer_conc	: Concentration of the primer [nM]. Template strand which concentration is typically very low and may be ignored and so not included in this function.
		-Na, K, Tris, Mg, dNTPs	: Concentration of the respective ions [mM]. If any of K, Tris, Mg and dNTPS is non-zero, a 'sodium-equivalent' concentration
								  is calculated and used for salt correction (von Ahsen et al., 2001).
		-ion_corr	: See method 'Tm_GC'. Default = True. (0 means no salt correction).	
	"""
	if not seq:
		raise ValueError('Please provide an input sequence!')
	seq = seq_qc(seq)
	if not compl_seq:											# If no complementary seq is provided, the complement of the given seq is caclulated and used
			compl_seq = complement(seq)
	compl_seq = seq_qc(compl_seq)
	
	temp_seq = str(seq)
	temp_compl_seq = str(compl_seq)
	dH			= 0											# dH stands for delta H
	dS			= 0											# dS stands for delta S
	dH_index	= 0											# dH_index stands for deta H_index in a table (here position 0)
	dS_index	= 1											# dS_index stands for delta S_index in a table (here position 1)
	left_tmm	= temp_compl_seq[:2][::-1] + '/' + temp_seq[:2][::-1]
	if left_tmm in DNA_TMM_table:
		dH += DNA_TMM_table[left_tmm][dH_index]
		dS += DNA_TMM_table[left_tmm][dS_index]
		temp_compl_seq	= temp_compl_seq[1:]
		temp_seq		= temp_seq[1:]
	# right_tmm	= temp_compl_seq[-2:] + '/' + temp_seq[-2:]
	right_tmm	= temp_seq[-2:] + '/' + temp_compl_seq[-2:] 
	if right_tmm in DNA_TMM_table:
		dH += DNA_TMM_table[right_tmm][dH_index]
		dS += DNA_TMM_table[right_tmm][dS_index]	
		temp_compl_seq	= temp_compl_seq[:-1]	
		temp_seq		= temp_seq[:-1]
	# General initiation value
	dH += DNA_NN_table['init'][dH_index]
	dS += DNA_NN_table['init'][dS_index]
	# delta H and delta S correction for  Duplex with no (allA/T) or at least one (oneG/C) GC pair coefficients need not be considered while using SantaLucia & Hicks (2004) model
	# delta H and delta S correction for 5' end being T need not be considered while using SantaLucia & Hicks (2004) model 	
	# delta H and delta S correction for A/T terminal basepairs (for the original seq and not for the end trimmed temp_seq)
	### CONSIDERS TERMINAL MISMATCHES WHILE COUNTING TERMINAL A/Ts 
	terminals	=	seq[0]+'/'+compl_seq[0] + seq[-1] +'/'+compl_seq[-1]
	count_AT	=	terminals.count('A/T')+terminals.count('T/A')
	dH += DNA_NN_table['init_A/T'][dH_index] * count_AT
	dS += DNA_NN_table['init_A/T'][dS_index] * count_AT	
	# delta H and delta S calculations for matches and mimatches
	for base_index in range(len(temp_seq)-1):
		NN	=	temp_seq[base_index:base_index+2] + '/' + temp_compl_seq[base_index:base_index+2]
		if NN in DNA_IMM_table:
			dH += DNA_IMM_table[NN][dH_index]
			dS += DNA_IMM_table[NN][dS_index]
		elif NN[::-1] in DNA_IMM_table:
			dH += DNA_IMM_table[NN[::-1]][dH_index]
			dS += DNA_IMM_table[NN[::-1]][dS_index]
		elif NN in DNA_NN_table:
			dH += DNA_NN_table[NN][dH_index]
			dS += DNA_NN_table[NN][dS_index]
		elif NN [::-1] in DNA_NN_table:
			dH += DNA_NN_table[NN[::-1]][dH_index]
			dS += DNA_NN_table[NN[::-1]][dS_index]		
	if ion_corr:
		seq_len	=	len(seq)
		correction_factor = ion_correction(Na, K, Tris, Mg, dNTPs, seq_len)
		dS 	+=	correction_factor
	x	=	4															#	x = 4 if not self complementary; x = 1 if self complementary
	R	=	1.9872  													#	Universal gas constant in Cal/degrees C*Mol
	primer_conc = primer_conc/1000000000.0								#	To convert nM into molar; do not multiply with 2 if the other strand is genomic DNA template, which can be negligible = 
	Tm	= 	(1000* dH) / (dS + (R * (math.log(primer_conc/x)))) - 273.15 
	return ("{0:.2f}".format(round(Tm,2)))								# Round to 2 decimals


if __name__ == "__main__":
    print('CTTCTGCAATGCCAAGTCCAG')
    print('full match: ', NN_Tm(seq='CTTCTGCAATGCCAAGTCCAG', compl_seq='GAAGACGTTACGGTTCAGGTC'))
    print('One mismatch', NN_Tm(seq='CTTCTGCAATGCCAAGTCCAG', compl_seq='GAAGACGTCACGGTTCAGGTC'))


### Updated: 07/12/2019
# rename "xrange" to "range" in order to fit python3
# Add __main__ to debug
# setting default parameters to: primer_conc=100, Na=0, K=50, Tris=10, Mg=1.5, dNTPs=0.2, ion_corr=True in order to debug

### Updated: version-1.05 06/29/2017
# "right_tmm" definition modified



import os

from flask import Flask, render_template

from . import db, run, download, config
from primerserver2.core import version

app = Flask(__name__)
@app.route('/')
def index():
    check = config.check()
    if check['status'] == 'error':
        return render_template('error.html', msg=check['msg'])
    else:
        return render_template('index.html', ver=version.get())

app.register_blueprint(db.bp)
app.register_blueprint(run.bp)
app.register_blueprint(download.bp)



#!/usr/bin/env python3

'''PrimerServer2: a high-throughput primer design and specificity-checking platform
Github: https://github.com/billzt/PrimerServer2
This is a CLI script for preparing the exon-exon junction database (in JSON format) for PrimerServer2
'''

import json
import argparse

from primerserver2.core.qRT import get_junction
from primerserver2.core import version

def main():
    parser = argparse.ArgumentParser(description='primertool-junctions: preparing the exon-exon junction database for PrimerServer2', \
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-v', '--version', action='version', version='%(prog)s '+version.get())
    parser.add_argument('template', help='the template file in FASTA format')
    parser.add_argument('gff', help='the annotation file in GFF3 format')
    parser.add_argument('-f', '--features', help='features in GFF3 (the 3rd column) that should be parsed as exons. \
        Multiple features separated by comma are allowed', default='exon')
    parser.add_argument('-i', '--id', help='the key in GFF3 (in the last column) that should be treated as the transcript ID for exons. \
        ', default='Parent')
    parser.add_argument('-r', '--remove-version', help='remove the version code (such as .1, .2, ...) from isoform IDs. \
        ', action='store_true')
    args = parser.parse_args()

    junction_data = get_junction(gff_file=args.gff, features=args.features.split(','), rna_print_ID_key=args.id, remove_version=args.remove_version)

    with open(args.template+'.junctions.json', 'w') as f:
        json.dump(junction_data, f, indent=4)

if __name__ == "__main__":
    main()



import json
import os

from flask import Blueprint, request, current_app, make_response

from primerserver2.web.config import load

bp = Blueprint('db', __name__)
@bp.route('/dbselect')
def dbselect():
    web_config = load()
    dbs_in_select = {}
    for (template, meta) in web_config['templates'].items():
        (desc, group, ids) = (meta['description'], meta['group'], meta['IDs'])
        if group not in dbs_in_select:
            dbs_in_select[group] = {}
        if template not in dbs_in_select[group]:
            dbs_in_select[group][template] = {}
        dbs_in_select[group][template]['desc'] = desc
        dbs_in_select[group][template]['IDs'] = ids
    return json.dumps(dbs_in_select, indent=4)

@bp.route('/dbinfo')
def dbinfo():
    web_config = load()
    return json.dumps(web_config['templates'], indent=4)
    
@bp.route('/dbdownload/<template>/')
def dbdownload(template):
    web_config = load()
    db_dir = web_config['templates_directory']
    result = os.popen(f'cut -f 1,2 {db_dir}/{template}.fai').read()
    response = make_response(result)
    response.headers["Content-Disposition"] = f"attachment; filename={template}.fai"
    response.headers["Content-type"] = "application/octet-stream"
    return response


import json
import os
import sys

from primerserver2.core import global_var

def sort_rank(primers, dbs, max_num_return=10, use_isoforms=False):
    '''
        sort primers within each site based on their amplicons number in the main db
        Input:
            primers: generated by run_blast
            dbs: used in run_blast

        Optional:
            max_num_return: only output this number of primers
        Return:
            updated primers with PRIMER_PAIR_AMPLICON_NUM_RANK_X attribute
    '''
    if global_var.stop_run is True:
        return {'error': 'Stop running'}
        
    main_db = os.path.basename(dbs[0])
    if use_isoforms is True:
        isoform_data = json.load(open(dbs[0]+'.isoforms.json'))

    for (id, primer) in primers.items():
        if use_isoforms is True:
            template = id.split('-')[0]
            isoforms = []
            if template in isoform_data:
                isoforms = isoform_data[template]

        rank_to_amplicon_valid_num = {}
        for rank in range(0, primer['PRIMER_PAIR_NUM_RETURNED']):
            if f'PRIMER_PAIR_{rank}_AMPLICONS' not in primer[main_db]:
                continue
            if use_isoforms is False:
                rank_to_amplicon_valid_num[rank] = len(primer[main_db][f'PRIMER_PAIR_{rank}_AMPLICONS'])
            else:
                rank_to_amplicon_valid_num[rank] = 1
                for (i, amplicon) in enumerate(primer[main_db][f'PRIMER_PAIR_{rank}_AMPLICONS']):
                    sseqid = amplicon['plus']['sseqid']
                    if sseqid not in isoforms:
                        rank_to_amplicon_valid_num[rank] += 1
                        primer[main_db][f'PRIMER_PAIR_{rank}_AMPLICONS'][i]['isoform'] = False
                    else:
                        primer[main_db][f'PRIMER_PAIR_{rank}_AMPLICONS'][i]['isoform'] = True

        
        num_output_primers = 0
        for (i, rank) in enumerate(sorted(rank_to_amplicon_valid_num, key=rank_to_amplicon_valid_num.__getitem__)):
            primers[id][f'PRIMER_PAIR_AMPLICON_NUM_RANK_{i}'] = rank
            num_output_primers += 1
            if num_output_primers==max_num_return:
                break
        
        primers[id]['PRIMER_PAIR_NUM_RETURNED_FINAL'] = num_output_primers
    
    return primers

if __name__ == "__main__":
    global_var.init()
    primers = json.load(open('tests/_internal_/run_blast.json'))
    dbs = ['example.fa']
    print(json.dumps(sort_rank(primers, dbs), indent=4))




import json
import sys
import os

def tsv(primers_dict, dbs):
    print_line = []
    header_line = '\t'.join(['#Site_ID', 'Primer_Rank', 'Primer_Seq_Left', 'Primer_Seq_Right', 'Primer_Seq_Oligo', 'Target_Amplicon_Size', \
        'Primer_Pair_Penalty_Score', 'Primer_Rank_in_Primer3_output', 'Tm_left', 'Tm_Right', 'Tm_Oligo', \
            'Database\tPossible_Amplicon_Number\tAmplicon_Size_in_DB\tAmplicon_Region_in_DB\t'*len(dbs)])
    print_line.append(header_line)
    for (id, primers) in primers_dict.items():
        template_id = id.split('-')[0]

        # No primers
        if primers['PRIMER_PAIR_NUM_RETURNED']==0:
            print_data = []
            print_data.append(id)
            print_data.append('No_Primer')
            if 'PRIMER_LEFT_EXPLAIN' in primers:
                print_data.append(primers['PRIMER_LEFT_EXPLAIN'].replace(' ', '_'))
                print_data.append(primers['PRIMER_RIGHT_EXPLAIN'].replace(' ', '_'))
                print_data.append(primers['PRIMER_PAIR_EXPLAIN'].replace(' ', '_'))
            print_line.append('\t'.join(print_data))
            print_line.append('###')
            continue

        if 'PRIMER_PAIR_NUM_RETURNED_FINAL' in primers:
            primer_num = primers['PRIMER_PAIR_NUM_RETURNED_FINAL']
        else:
            primer_num = primers['PRIMER_PAIR_NUM_RETURNED']
        for amplicon_rank in range(0, primer_num):
            print_data = []
            print_data.append(id)
            print_data.append(str(amplicon_rank))
            if 'PRIMER_PAIR_NUM_RETURNED_FINAL' in primers:
                raw_rank = primers[f'PRIMER_PAIR_AMPLICON_NUM_RANK_{amplicon_rank}']
            else:
                raw_rank = amplicon_rank
            print_data.append(primers[f'PRIMER_LEFT_{raw_rank}_SEQUENCE'])
            print_data.append(primers[f'PRIMER_RIGHT_{raw_rank}_SEQUENCE'])
            if f'PRIMER_INTERNAL_{raw_rank}_SEQUENCE' in primers:
                print_data.append(primers[f'PRIMER_INTERNAL_{raw_rank}_SEQUENCE'])
            else:
                print_data.append('NA')
            print_data.append(str(primers[f'PRIMER_PAIR_{raw_rank}_PRODUCT_SIZE']))
            print_data.append(str(round(primers[f'PRIMER_PAIR_{raw_rank}_PENALTY'],2)))
            print_data.append(str(raw_rank))
            print_data.append(str(round(primers[f'PRIMER_LEFT_{raw_rank}_TM'], 2)))
            print_data.append(str(round(primers[f'PRIMER_RIGHT_{raw_rank}_TM'], 2)))
            if f'PRIMER_INTERNAL_{raw_rank}_TM' in primers:
                print_data.append(str(round(primers[f'PRIMER_INTERNAL_{raw_rank}_TM'], 2)))
            else:
                print_data.append('NA')
            for db in dbs:
                db_base = os.path.basename(db)
                print_data.append(db_base)
                if db_base in primers:
                    if f'PRIMER_PAIR_{raw_rank}_AMPLICONS' in primers[db_base]:
                        amplicon_num_in_db = 0
                        for amplicon in primers[db_base][f'PRIMER_PAIR_{raw_rank}_AMPLICONS']:
                            if 'isoform' not in amplicon:
                                amplicon_num_in_db+=1
                            elif amplicon['isoform'] is False:
                                amplicon_num_in_db+=1
                            elif amplicon['plus']['sseqid'] == template_id:
                                amplicon_num_in_db+=1

                        print_data.append(str(amplicon_num_in_db))
                        if amplicon_num_in_db==1:
                            print_data.append(str(primers[db_base][f'PRIMER_PAIR_{raw_rank}_AMPLICONS'][0]['product_size']))
                            print_data.append(str(primers[db_base][f'PRIMER_PAIR_{raw_rank}_AMPLICONS'][0]['region']))
                        else:
                            print_data.extend(('NA', 'NA'))
                    else:
                        print_data.extend(('NA', 'NA', 'NA'))
                else:
                    print_data.extend(('NA', 'NA', 'NA'))
            

            print_line.append('\t'.join(print_data))
        print_line.append('###')

    return '\n'.join(print_line)

def dimer_list(dimers):
    print_line = ['################### Dimers #################']
    if 'error' in dimers:
        print_line.append('ERROR: '+dimers['error'])
    else:
        header_line = '\t'.join(['#Site_1', 'Primer_Rank_1', 'Seq_1', 'Site_2', 'Primer_Rank_2', 'Seq_2', 'Tm'])
        print_line.append(header_line)
        for site_1 in dimers.keys():
            for primer_1 in dimers[site_1].keys():
                for site_2 in dimers[site_1][primer_1].keys():
                    for primer_2 in dimers[site_1][primer_1][site_2].keys():
                        dimer = dimers[site_1][primer_1][site_2][primer_2]
                        print_line.append('\t'.join([site_1, str(primer_1), dimer['seq_1'], \
                            site_2, str(primer_2), dimer['seq_2'], str(dimer['Tm'])]))
            print_line.append('###')
    return '\n'.join(print_line)


if __name__ == "__main__":
    primers = json.load(open('tests/_internal_/sort_primers2.json'))
    dbs = ["Ghir.JGI.genomic", "Gbar.IGDB.genomic"]
    print(tsv(primers, dbs))



import json

from flask import Blueprint, request

from primerserver2.core import output

bp = Blueprint('download', __name__)
@bp.route('/download_tsv', methods=['POST'])
def download_tsv():
    json_data = json.loads(request.form['json'])
    dbs = request.form['dbs'].split(',')
    print_lines = output.tsv(json_data['primers'], dbs)
    return print_lines




import re
import os
import json
import subprocess

from primerserver2.core import global_var

def get_template_len(template_file):
    '''
        return a dict of each seq's len based on the faidx
    '''
    template_len_for = {}
    with open(template_file+'.fai') as f:
        for line in f:
            (id, length) = line.split('\t')[0:2]
            template_len_for[id] = int(length)
    return template_len_for

def faidx(template_file, region_string):
    '''
        return a list of sequence dict using samtools faidx based on sequence file (fa)
        and region_string: 
            seq1:1-100
            seq2:2-200
            ...
    '''
    if os.path.isfile(template_file)==False:
        raise Exception(f'File {template_file} is not found')

    cmd = f'samtools faidx -r - {template_file}'
    p = subprocess.run(cmd, input=region_string, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, shell=True, encoding='ascii')
    seqs = p.stdout.split('>')
    del seqs[0]
    result_seqs = {}
    for seq in seqs:
        seq_split = seq.splitlines()
        result_seqs[seq_split[0]] = ''.join(seq_split[1:])
    return result_seqs

def read_junction(template_file):
    if os.path.isfile(f'{template_file}.junctions.json') is True:
        return json.load(open(f'{template_file}.junctions.json'))
    else:
        return {}

def get_junction_poses(junction_data, id):
    if id not in junction_data:
        return []
    else:
        return [x['pos'] for x in junction_data[id]['junctions'] if x['intron_len']>100]


def judge_input_type(query):
    '''
        whether input by pos or by seq?
        return: pos or seq
    '''
    line1 = query.splitlines()[0]
    data1 = re.split(r'\s+', line1.strip())[0]
    if re.search(r'[^ATGCNatgcn\[\]]', data1) is None and len(data1)>40:
        return 'seq'
    else:
        return 'pos'

def build_by_seq(query, primer_type, primer_num_return=30, size_min=70, \
    size_max=1000, pick_internal=False, Tm_opt=60):
    '''
        Input:
            query: a string in multi-lines
        Return:
            a list of primer sites dict that can be passed to the design_primer module
            [{
                'id': , 
                'template': ,
                'type': ,
                'pos': ,
                'length':,
                'size_min':,
                'size_max':,
                'primer_num_return':,
            }]
    '''
    if global_var.stop_run is True:
        return {'error': 'Stop running'}

    primer_sites = []
    primer_site_rank = 0
    for line in query.splitlines():
        if line.strip()=='':
            continue
        primer_site_rank += 1
        seq = re.split(r'\s+', line.strip())[0]

        # SEQUENCE_TARGET, FORCE_END
        if primer_type=='SEQUENCE_TARGET' or primer_type=='FORCE_END':
            if '[' not in seq or ']' not in seq:
                return {'error': f'seq {seq} has no [target regions]'}
            if seq.count('[')>1 or seq.count(']')>1:
                return {'error': f'multiple [target regions] in {seq}'}
            pos = seq.find('[')
            length = seq.find(']')-seq.find('[')
            primer_sites.append({
                'id': 'S'+str(primer_site_rank)+'-'+str(pos)+'-'+str(length), 
                'template': seq.replace('[','').replace(']',''),
                'type': primer_type,
                'pos': pos,
                'length':length,
                'size_min':size_min,
                'size_max':size_max,
                'primer_num_return':primer_num_return,
                'pick_internal': pick_internal,
                'junction': [],
                'Tm_opt': Tm_opt
            })
        # SEQUENCE_INCLUDED_REGION
        if primer_type=='SEQUENCE_INCLUDED_REGION':
            if '[' not in seq and ']' not in seq:
                pos = 1
                length = len(seq)-1
            else:
                pos = seq.find('[')
                length = seq.find(']')-seq.find('[')
            primer_sites.append({
                'id': 'S'+str(primer_site_rank)+'-'+str(pos)+'-'+str(length), 
                'template': seq.replace('[','').replace(']',''),
                'type': primer_type,
                'pos': pos,
                'length':length,
                'size_min':size_min,
                'size_max':size_max,
                'primer_num_return':primer_num_return,
                'pick_internal': pick_internal,
                'junction': [],
                'Tm_opt': Tm_opt
            })
    return primer_sites



def build_by_pos(query, template_file, primer_type, primer_num_return=30, size_min=70, \
    size_max=1000, pick_internal=False, use_junction=False, Tm_opt=60):
    '''
        Input:
            query: a string in multi-lines
        Return:
            a list of primer sites dict that can be passed to the design_primer module
            [{
                'id': , 
                'template': ,
                'type': ,
                'pos': ,
                'length':,
                'size_min':,
                'size_max':,
                'primer_num_return':,
            }]
    '''
    if global_var.stop_run is True:
        return {'error': 'Stop running'}
        
    template_len_for = get_template_len(template_file)
    primer_sites = []
    retrieve_region2raw_region = {}
    for line in query.splitlines():
        if line.strip()=='':
            continue
        query_data = re.split(r'\s+', line.strip())
        if query_data[0]=='':
            return {'error': f'Your input {line.strip()} has no template seqs'}
        (chr, pos, length) = (query_data[0], -1, -1)
        if chr not in template_len_for:
            return {'error': f'Your input {line.strip()} has no template seqs'}
        chr_len = template_len_for[chr]
        if len(query_data)>1 :     # pos is provided
            pos = int(query_data[1].replace(',', ''))
            if pos<0 or pos>chr_len:
                return {'error': f'Your input {line.strip()} has wrong pos: {pos}: it must be between 1 and {chr_len}'}
        if len(query_data)>2 :      # len is provided
            length = int(query_data[2])
            if pos+length > chr_len:
                return {'error': f'Your input {line.strip()} has wrong region: {pos}+{length}: it cannot be exceed {chr_len}'}
        if len(query_data)>3 :
            size_min = int(query_data[3])
            size_max = int(query_data[4])
        
        # some adjustion
        if primer_type=='SEQUENCE_TARGET' or primer_type=='FORCE_END':
            if pos==-1:
                return {'error': f'Your input {line.strip()} lack position, which is required by {primer_type}'}
            if length==-1:
                length=1
        if primer_type=='SEQUENCE_INCLUDED_REGION':
            if pos==-1:
                pos=1
            if length==-1:
                length=chr_len-pos

        # retrieve_start and retrieve_end are used by samtools to extract template sequences
        retrieve_start = max(pos-size_max, 1)
        retrieve_end = pos+length+size_max
        retrieve_region2raw_region[f'{chr}:{retrieve_start}-{retrieve_end}'] = [chr, pos, length, size_min, size_max, retrieve_start]
    
    if (len(retrieve_region2raw_region)==0):
        return {'error': f'Your input has no template seqs'}
    retrieve_region_string = '\n'.join(retrieve_region2raw_region.keys())
    result_seqs = faidx(template_file, retrieve_region_string)
    if use_junction is True:
        junction_data = read_junction(template_file)
    for (result_seq_id, result_seq) in result_seqs.items():
        (chr, pos, length, size_min, size_max, retrieve_start) = retrieve_region2raw_region[result_seq_id]
        if result_seq=='':
            return {'error': f'Your input: seq={chr} pos={pos} length={length} has no template seqs'}
        junctions = []
        if use_junction is True:
            junctions = get_junction_poses(junction_data, chr)
        primer_sites.append({
            'id': chr+'-'+str(pos)+'-'+str(length), 
            'template': result_seq,
            'type': primer_type,
            'pos': pos-retrieve_start,
            'length':length,
            'size_min':size_min,
            'size_max':size_max,
            'primer_num_return':primer_num_return,
            'pick_internal': pick_internal,
            'junction': junctions,
            'Tm_opt': Tm_opt
        })
    
    return primer_sites

if __name__ == "__main__":
    global_var.init()
    import argparse
    parser = argparse.ArgumentParser(description='Your title here')
    parser.add_argument('query_design', help='input file', type=argparse.FileType('r'))
    parser.add_argument('type', help='SEQUENCE_TARGET, SEQUENCE_INCLUDED_REGION, FORCE_END')
    args = parser.parse_args()
    query = args.query_design.read()
    input_type = judge_input_type(query)
    if input_type=='pos':
        primer_sites = build_by_pos(query=query, template_file='tests/example.fa', primer_type=args.type, primer_num_return=30)
    else:
        primer_sites = build_by_seq(query=query, primer_type=args.type, primer_num_return=30)
    print(json.dumps(primer_sites, indent=4))



import re
import json
import sys

from primerserver2.core.make_sites import faidx
from primerserver2.core.Santalucia_NN_Tm import complement, rev_complement, NN_Tm, transform_degenerate
from primerserver2.core.make_primers import make_primers

def filter_len(blast_out, len_min, len_max, chr_main_only=True):    # 1s
    '''
        return 
            'amplicons': 
                primer_id=>
                    primer_rank=> [hit-pairs]
            'regions_primer': a region string for each primer used by samtools faidx;
    '''

    # collect hits data
    primer_hits_all_sites = {}
    for line in blast_out.splitlines():
        (qseqid, qlen, qstart, qend, sseqid, slen, sstart, send, sstrand) = line.strip().split('\t')
        
        # include main chrs only
        # if 'random' in sseqid or 'alt' in sseqid or 'Un' in sseqid or 'scaffold' in sseqid:
        #     continue

        # collect
        # qseqid: {id}.{rank}.L or {id}.{rank}.R
        qseqid_data = qseqid.split('.')
        primer_rank = qseqid_data[-2]
        primer_id = '.'.join(qseqid_data[0:-2])
        if primer_id not in primer_hits_all_sites:
            primer_hits_all_sites[primer_id] = {}
        if primer_rank not in primer_hits_all_sites[primer_id]:
            primer_hits_all_sites[primer_id][primer_rank] = {}
        if sseqid not in primer_hits_all_sites[primer_id][primer_rank]:
            primer_hits_all_sites[primer_id][primer_rank][sseqid] = []
        primer_hits_all_sites[primer_id][primer_rank][sseqid].append({
            'qseqid': qseqid,
            'qlen': int(qlen),
            'qstart': int(qstart),
            'qend': int(qend),
            'slen': int(slen),
            'sstart': int(sstart),
            'send': int(send),
            'sstrand': sstrand
        })
    
    # # filter length between plus and minus
    amplicons = {}
    hits_regions = {}   # make region file (FASTA)

    for primer_id in primer_hits_all_sites.keys():
        amplicons[primer_id] = {}
        for primer_rank in primer_hits_all_sites[primer_id].keys():
            for sseqid in primer_hits_all_sites[primer_id][primer_rank].keys():
                primer_hits = sorted(primer_hits_all_sites[primer_id][primer_rank][sseqid], key=lambda i: i['sstart'])
                primer_hits_plus = [x for x in primer_hits if x['sstrand']=='plus']
                primer_hits_minus = [x for x in primer_hits if x['sstrand']=='minus']
                for hit_plus in primer_hits_plus: 
                    for hit_minus in primer_hits_minus:
                        size = hit_minus['sstart']-hit_plus['sstart']
                        if size>len_min and size<len_max:
                            # extend this pair:
                            # plus
                            start_plus = hit_plus['sstart']-(hit_plus['qstart']-1)
                            if start_plus<1:    # extension failed
                                continue
                            end_plus = hit_plus['send']+(hit_plus['qlen']-hit_plus['qend'])
                            if end_plus>hit_plus['slen']: # extension failed
                                continue
                            
                            # minus
                            start_minus = hit_minus['sstart']+(hit_minus['qstart']-1)
                            if start_minus>hit_minus['slen']: # extension failed
                                continue
                            end_minus = hit_minus['send']-(hit_minus['qlen']-hit_minus['qend'])
                            if end_minus<1:  # extension failed
                                continue

                            # store this pair
                            if primer_rank not in amplicons[primer_id]:
                                amplicons[primer_id][primer_rank] = []
                            amplicons[primer_id][primer_rank].append({
                                'plus': {
                                    'qseqid': hit_plus['qseqid'], # LEFT or RIGHT
                                    'sseqid': sseqid,
                                    'sstart': start_plus,
                                    'send': end_plus
                                },
                                'minus': {
                                    'qseqid': hit_minus['qseqid'], # LEFT or RIGHT
                                    'sseqid': sseqid,
                                    'sstart': end_minus,    # flip
                                    'send': start_minus
                                }
                            })

                            # store regions
                            hits_regions[sseqid+':'+str(start_plus)+'-'+str(end_plus)] = 1
                            hits_regions[sseqid+':'+str(end_minus)+'-'+str(start_minus)] = 1 # flip

    return {'amplicons':amplicons, 'regions_primer': '\n'.join(hits_regions.keys()) }

def filter_Tm(amplicons, query_primer_seq, hits_seqs, Tm_diff=20, max_amplicons=10, use_3_end=False):
    '''
        Input:
            query_primer_seq:
                "LEFT": XXXXXXXXXXXXXX
                "RIGHT": XXXXXXXXXXXX
    '''
    amplicons_filter = {}
    for primer_id in amplicons.keys():
        amplicons_filter[primer_id] = {}
        for primer_rank in amplicons[primer_id]:
            for amplicon in amplicons[primer_id][primer_rank]:
                # hit seqs (template)
                region_plus = amplicon['plus']['sseqid']+':'+str(amplicon['plus']['sstart'])+'-'+str(amplicon['plus']['send'])
                region_minus = amplicon['minus']['sseqid']+':'+str(amplicon['minus']['sstart'])+'-'+str(amplicon['minus']['send'])
                hit_seq_plus = hits_seqs[region_plus]
                hit_seq_minus = rev_complement(hits_seqs[region_minus])  # samtools faidx hasn't do rev_complement yet

                # query primers (query)
                query_primer_plus = query_primer_seq[amplicon['plus']['qseqid']]  # LEFT or RIGHT
                query_primer_minus = query_primer_seq[amplicon['minus']['qseqid']]  # LEFT or RIGHT

                # perfect Tm for this primer pair
                Tm_perfect = min(float(NN_Tm(seq=transform_degenerate(query_primer_plus), \
                    compl_seq=complement(transform_degenerate(query_primer_plus)))), \
                    float(NN_Tm(seq=transform_degenerate(query_primer_minus), \
                        compl_seq=complement(transform_degenerate(query_primer_minus)))))

                # 3' end in this amplicon
                if use_3_end is True:
                    # plus
                    if query_primer_plus[-1].upper() != hit_seq_plus[-1].upper():
                        continue
                    # minus
                    if query_primer_minus[-1].upper() != hit_seq_minus[-1].upper():
                        continue

                # Tm in this amplicon
                # plus
                if len(query_primer_plus) != len(hit_seq_plus):
                    raise Exception(f'Not the same length: \n \
                        Primer: {query_primer_plus} \n \
                        Hit: {region_plus} {hit_seq_plus}')
                Tm_plus = float(NN_Tm(seq=transform_degenerate(query_primer_plus), \
                    compl_seq=complement(transform_degenerate(hit_seq_plus))))
                if Tm_plus<Tm_perfect-Tm_diff:
                    continue

                # minus
                if len(query_primer_minus) != len(hit_seq_minus):
                    raise Exception(f'Not the same length: \n \
                        Primer: {query_primer_minus} \n \
                        Hit: {region_minus} {hit_seq_minus}')
                Tm_minus = float(NN_Tm(seq=transform_degenerate(query_primer_minus), \
                    compl_seq=complement(transform_degenerate(hit_seq_minus))))
                if Tm_minus<Tm_perfect-Tm_diff:
                    continue

                # store them
                amplicon['plus']['Tm'] = Tm_plus
                amplicon['plus']['qseq'] = query_primer_plus
                amplicon['plus']['sseq'] = hit_seq_plus
                amplicon['minus']['Tm'] = Tm_minus
                amplicon['minus']['qseq'] = query_primer_minus
                amplicon['minus']['sseq'] = hit_seq_minus
                amplicon['Tm'] = min(Tm_plus, Tm_minus)
                amplicon['region'] = amplicon['plus']['sseqid']+':'+str(amplicon['plus']['sstart'])\
                    +'-'+str(amplicon['minus']['send'])
                amplicon['product_size'] = amplicon['minus']['send']-amplicon['plus']['sstart']+1
                amplicon['product_seq'] = ''
                if primer_rank not in amplicons_filter[primer_id]:
                    amplicons_filter[primer_id][primer_rank] = []
                amplicons_filter[primer_id][primer_rank].append(amplicon)

                if len(amplicons_filter[primer_id][primer_rank])==max_amplicons:
                    break

    return amplicons_filter

def add_amplicon_seq(amplicons, template_file):
    amplicon_regions = ''
    for primer_id in amplicons.keys():
        for primer_rank in amplicons[primer_id]:
            for amplicon in amplicons[primer_id][primer_rank]:
                amplicon_regions += amplicon['region']+'\n'
    amplicon_seqs = faidx(template_file=template_file, region_string=amplicon_regions)
    for primer_id in amplicons.keys():
        for primer_rank in amplicons[primer_id]:
            for (i, amplicon) in enumerate(amplicons[primer_id][primer_rank]):
                amplicons[primer_id][primer_rank][i]['product_seq'] = amplicon_seqs[amplicon['region']]
    return amplicons

def add_isoform_annotation(amplicons, dbs):
    isoform_data = json.load(open(dbs[0]+'.isoforms.json'))
    for primer_id in amplicons.keys():
        for primer_rank in amplicons[primer_id]:
            for (i, amplicon) in enumerate(amplicons[primer_id][primer_rank]):
                sseqid = amplicon['plus']['sseqid']
                qseqid = amplicon['plus']['qseqid'].split('-')[0]
                if qseqid in isoform_data and sseqid in isoform_data[qseqid]:
                    amplicons[primer_id][primer_rank][i]['isoform'] = True
    return amplicons


if __name__ == "__main__":
    blast_out = open('tests/_internal_/query_blast.fa.out.large').read()
    amplicons = filter_len(blast_out=blast_out, len_min=75, len_max=1000)
    # filter_len(blast_out=blast_out, len_min=75, len_max=1000)
    # hits_seqs = faidx(template_file='tests/example.fa', region_string=amplicons['regions_primer'])
    hits_seqs = faidx(template_file='tests/_internal_/Ghir.JGI.genomic', region_string=amplicons['regions_primer'])
    # query_primer_seq={'P1.0.L':'CTTCTGCAATGCCAAGTCCAG',\
    #         'P1.0.R': 'GTGGTGAAGGGTCGGTTGAA'}
    query_primer_seq_lines = open('tests/_internal_/query_blast.fa.out.large.primer.seq').read().splitlines()
    i = iter(query_primer_seq_lines)
    query_primer_seq = dict(zip(i, i))
    report_amplicons = filter_Tm(amplicons['amplicons'], query_primer_seq=query_primer_seq, hits_seqs=hits_seqs)
    #report_amplicons = add_amplicon_seq(amplicons=report_amplicons, template_file='tests/example.fa')
    report_amplicons = add_amplicon_seq(amplicons=report_amplicons, template_file='tests/_internal_/Ghir.JGI.genomic')
    print(json.dumps(report_amplicons, indent=4))



import re
import json
import os
import sys
import subprocess
import multiprocessing as mp
import time

import primer3
import progressbar

from primerserver2.core.analysis_blast import filter_len, filter_Tm, add_amplicon_seq, add_isoform_annotation
from primerserver2.core.make_sites import faidx
from primerserver2.core.make_primers import make_primers
from primerserver2.core import global_var

def run_blast(p3_inputs):
    '''
        run blastn for a single primer pair, then use analysis_blast.py to obtain amplicons within Tm thredsholds
    '''
    if global_var.stop_run is True:
        return {'db': '', 'amplicons': {}}

    if len(p3_inputs)==0 :  # unfortunately, these sites have no primers
        return {'db': '', 'amplicons': {}}
    db = p3_inputs[0]['db']  # dbs are the same
    blast_query = ''
    query_primer_seq_dict = {}  # for the func: filter_Tm
    for x in p3_inputs:
        id = x['id']
        rank = x['rank']
        seq_L = x['seq_L']
        seq_R = x['seq_R']
        blast_query += f'>{id}.{rank}.L\n{seq_L}\n>{id}.{rank}.R\n{seq_R}\n'
        query_primer_seq_dict[f'{id}.{rank}.L'] = seq_L
        query_primer_seq_dict[f'{id}.{rank}.R'] = seq_R

    if os.path.isfile(db+'.nhr')==False and os.path.isfile(db+'.nal')==False:
        raise Exception(f'The database file is not complete: file {db}.nhr or {db}.nal is not found')
    cmd = f'blastn -task blastn-short -db {db} -evalue 30000 -word_size 7 ' \
        + '-perc_identity 60 -dust no -reward 1 -penalty -1 -max_hsps 500 -ungapped ' \
        + ' -outfmt "6 qseqid qlen qstart qend sseqid slen sstart send sstrand"'
    blast_out = subprocess.run(cmd, input=blast_query, stdout=subprocess.PIPE, shell=True, encoding='ascii').stdout
    amplicons = filter_len(blast_out=blast_out, len_min=p3_inputs[0]['checking_size_min'], \
        len_max=p3_inputs[0]['checking_size_max'])
    hits_seqs = faidx(template_file=db, region_string=amplicons['regions_primer'])
    report_amplicons = filter_Tm(amplicons['amplicons'], query_primer_seq=query_primer_seq_dict, \
        hits_seqs=hits_seqs, Tm_diff=p3_inputs[0]['Tm_diff'], use_3_end=p3_inputs[0]['use_3_end'], \
            max_amplicons=p3_inputs[0]['max_amplicon'])
    if p3_inputs[0]['report_amplicon_seq'] is True:
        report_amplicons = add_amplicon_seq(amplicons=report_amplicons, template_file=db)
    return {'db': os.path.basename(db), 'amplicons': report_amplicons}

def run_blast_parallel(primers, dbs, cpu=2, checking_size_min=70, checking_size_max=1000, \
    report_amplicon_seq=False, Tm_diff=20, use_3_end=False, monitor=True, max_amplicon=10):
    if global_var.stop_run is True:
        return {'error': 'Stop running'}

    # distribute tasks
    pool = mp.Pool(processes=cpu)
    multi_res = []
    global_var.all_tasks_num = 0
    global_var.current_task = 'blast'
    for db in dbs:
        p3_inputs = []
        for (id, primer) in primers.items():
            primer_num = primer['PRIMER_PAIR_NUM_RETURNED']
            for rank in range(0, primer_num):
                p3_inputs.append({
                    'id': id,
                    'rank': rank,
                    'db': db,
                    'seq_L': primer[f'PRIMER_LEFT_{rank}_SEQUENCE'],
                    'seq_R': primer[f'PRIMER_RIGHT_{rank}_SEQUENCE'],
                    'checking_size_min': checking_size_min,
                    'checking_size_max': checking_size_max,
                    'report_amplicon_seq': report_amplicon_seq,
                    'Tm_diff': Tm_diff,
                    'use_3_end': use_3_end,
                    'max_amplicon': max_amplicon
                })
        for i in range(0, int(len(p3_inputs)/5)+1):
            sub_start = i*5
            sub_end = min(len(p3_inputs), sub_start+5)
            global_var.all_tasks_num += 5
            multi_res.append(pool.apply_async(run_blast, (p3_inputs[sub_start:sub_end],)))

    # monitor
    if monitor is True:
        widgets = ['Checking specificity: ', progressbar.Counter(),\
            ' Finished', ' (', progressbar.Percentage(), ')', \
                progressbar.Bar(), progressbar.ETA()]
        bar = progressbar.ProgressBar(widgets=widgets, max_value=global_var.all_tasks_num).start()

    while True:
        global_var.complete_count = sum([1 for x in multi_res if x.ready()])*5
        if global_var.complete_count == global_var.all_tasks_num:
            if monitor is True:
                bar.finish()
            break
        if monitor is True:
            bar.update(global_var.complete_count)
        time.sleep(1)
        
    # Results
    for result in multi_res:
        result_data = result.get()  # db and amplicons
        db = result_data['db']
        if db== '': # unfortunately, these sites have no primers
            continue
        amplicons = result_data['amplicons']
        for id in amplicons.keys():
            if db not in primers[id]:
                primers[id][db] = {}
            for rank in amplicons[id].keys():
                primers[id][db][f'PRIMER_PAIR_{rank}_AMPLICONS'] = amplicons[id][rank]

    return primers

if __name__ == "__main__":
    global_var.init()
    with open('tests/query_check_multiple') as f:
        primers = make_primers(query=f.read())
        dbs = ['tests/example.fa']
        report_primers = run_blast_parallel(primers, dbs, cpu=10, report_amplicon_seq=True)
        print(json.dumps(report_primers, indent=4))



#!/usr/bin/env python3

'''PrimerServer2: a high-throughput primer design and specificity-checking platform
Github: https://github.com/billzt/PrimerServer2
This is a CLI script for PrimerServer2
'''

import argparse
import re
import os
import sys
import json
import shutil

from distutils.version import LooseVersion

from primerserver2.core import make_sites, make_primers, design_primer, run_blast, sort_primers, output, analysis_blast, global_var
from primerserver2.core import multiplex
from primerserver2.core import version

def make_args():
    parser = argparse.ArgumentParser(description='primertool: the command-line version of PrimerServer2', \
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-v', '--version', action='version', version='%(prog)s '+version.get())

    # parent arguments
    # These arguments are used by all the three modes
    parent_parser_all = argparse.ArgumentParser(add_help=False)
    group_io = parent_parser_all.add_argument_group('Input (REQUIRED)')
    group_io.add_argument('query', help='query file. (STDIN is acceptable)', \
        type=argparse.FileType('r'))
    group_io.add_argument('templates', help='template file in FASTA format. \
        Allowing multiple files (separated by comma), where the first one is used to design primers and/or order the primer \
            specificity')

    group_all = parent_parser_all.add_argument_group('Overall Setttings')
    group_all.add_argument('--primer-num-retain', type=int, help='The maximum number of primers to retain in each \
        site in the final report.', default=10)
    group_all.add_argument('--check-multiplex', help='Checking dimers between primers in different sites, which is useful in \
        multiplex PCR.', action='store_true')
    group_all.add_argument('--Tm-diff', type=int, help='The mininum difference of melting temperature (℃) \
        suggested to produce off-target amplicon or primer dimers. Suggest >10', default=20)
    group_all.add_argument('-p', '--cpu', type=int, help='Used CPU number.', default=2)
    group_all.add_argument('-o', '--out', help="Output primers in JSON format. default: {query}.out", \
        type=argparse.FileType('w'))
    group_all.add_argument('-t', '--tsv', help="Output primers in TSV format. default: {query}.tsv", \
        type=argparse.FileType('w'))

    # These arguments are used by design and full
    parent_parser_design = argparse.ArgumentParser(add_help=False)
    group_design = parent_parser_design.add_argument_group('Design Primers')
    group_design.add_argument('--type', choices=['SEQUENCE_TARGET', 'SEQUENCE_INCLUDED_REGION', 'FORCE_END'],\
        help='designing primer types', default='SEQUENCE_TARGET')
    group_design.add_argument('--pick-oligo', action='store_true', help='Pick internal Oligos (Probes) for qRT-PCR')
    group_design.add_argument('--product-size-min', type=int, help='Lower limit of the product amplicon size range (bp).', \
        default=70)
    group_design.add_argument('--product-size-max', type=int, help='Upper limit of the product amplicon size range (bp).', \
        default=1000)
    group_design.add_argument('--Tm-opt', type=int, help='Optmized melting temperature for primers (℃).', \
        default=60)
    group_design.add_argument('--primer-num-return', type=int, help='The maximum number of primers to return in Primer3 \
        designing results.', default=30)
    # qRT-PCR specific
    group_design.add_argument('--junction', help='Primer pair must be separated by at least one intron on the \
        corresponding genomic DNA; or primers must span an exon-exon junction. Junction data in JSON format should be prepared by \
            the command-line primertool-junctions', action='store_true')
    
    # These arguments are used by check and full
    parent_parser_check = argparse.ArgumentParser(add_help=False)
    group_check = parent_parser_check.add_argument_group('Check Specificity')
    group_check.add_argument('-3', '--use-3-end', help='If turned on, primer pairs having at least one mismatch at the 3 end\
        position with templates would not be considered to produce off-target amplicon, even if their melting temperatures \
            are high. Turn on this would find more candidate primers, but might also have more false positives\
                ', action='store_true')
    group_check.add_argument('--checking-size-min', type=int, help='Lower limit of the checking amplicon size range (bp).', \
        default=50)
    group_check.add_argument('--checking-size-max', type=int, help='Upper limit of the checking amplicon size range (bp).', \
        default=2000)
    group_check.add_argument('--amplicon-num-max', type=int, help='The maximum number of amplicons for checking.', \
        default=10)
    group_check.add_argument('-a', '--report-amplicon-seqs', help="Get amplicon seqs (might be slow)", action='store_true')
    # qRT-PCR specific
    group_check.add_argument('--isoform', help="Allow primers targeting on alternative isoforms and still regard them \
        as specific ones. Isoform data in JSON format should be prepared by the command-line primertool-isoforms", action='store_true')

    # sub commands
    subparsers = parser.add_subparsers(help='Sub commands: running mode')
    subparsers.required = True
    subparsers.dest = 'run_mode'
    # sub commands: full (design primers and check specificity)
    parser_full = subparsers.add_parser('full', help='design primers, check their specificity and output specific ones', \
        parents=[parent_parser_all, parent_parser_design, parent_parser_check], \
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    # sub commands: design (design primers only, don't check specificity)
    parser_design = subparsers.add_parser('design', help="design primers only, don't check specificity", \
        parents=[parent_parser_all, parent_parser_design], formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    # sub commands: check (check specificity only, skip the design step)
    parser_check = subparsers.add_parser('check', help='check specificity only, skip the design step', \
        parents=[parent_parser_all, parent_parser_check], formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    args = parser.parse_args()
    return args

def check_environments(args):
    if shutil.which('samtools') is None:
        raise Exception('No samtools detected in your system')

    samtools_version = os.popen('samtools --version').readlines()[0].strip().split(' ')[1]
    if LooseVersion(samtools_version) < LooseVersion('1.9'):
        raise Exception(f'Your samtools version is v{samtools_version}, but >=v1.9 is required')

    if shutil.which('blastn') is None:
        raise Exception('No NCBI-BLAST+ (blastn) detected in your system')

    if shutil.which('makeblastdb') is None:
        raise Exception('No NCBI-BLAST+ (makeblastdb) detected in your system')

def check_templates(args):
    for (i, template) in enumerate(args.templates.split(',')):
        if os.path.isfile(template) is False:
            raise Exception(f'File not found: {template}')
        if os.path.isfile(template+'.fai') is False:
            code = os.system(f'samtools faidx {template} 2>/dev/null')
            if code != 0:
                raise Exception(f'File {template} cannot be indexed by samtools faidx. Perhaps it is not in FASTA format')
        if os.path.isfile(template+'.nhr') is False and os.path.isfile(template+'.nal') is False:
            code = os.system(f'makeblastdb -dbtype nucl -in {template} 2>/dev/null')
            if code != 0:
                raise Exception(f'File {template} cannot be indexed by makeblastdb.')
        
        # qPCR specific
        if i==0:
            if os.path.isfile(template+'.junctions.json') is False and 'junction' in args and args.junction is True:
                raise Exception(f'The junction file for {template} is not ready. Parameter --junction is not allowed')
            if os.path.isfile(template+'.isoforms.json') is False and 'isoform' in args and args.isoform is True:
                raise Exception(f'The isoform file for {template} is not ready. Parameter --isoform is not allowed')

def check_qPCR(args):
    if args.run_mode!='check' and args.junction is True and args.type!='SEQUENCE_INCLUDED_REGION':
        raise Exception('Parameter --junction must be used with --type=SEQUENCE_INCLUDED_REGION')

def run(args):
    ###################  Design primers ###################
    query_string = args.query.read()
    dbs = args.templates.split(',')
    if args.run_mode=='check':
        primers = make_primers.make_primers(query=query_string)
        if 'error' in primers:
            raise Exception(primers['error'])
    else:
        if args.run_mode=='design':
            primer_num_return = args.primer_num_retain
        else:
            primer_num_return = args.primer_num_return
        if make_sites.judge_input_type(query_string)=='pos':
            sites = make_sites.build_by_pos(query=query_string, template_file=dbs[0], primer_type=args.type, \
                primer_num_return=primer_num_return, size_min=args.product_size_min, size_max=args.product_size_max, \
                    pick_internal=args.pick_oligo, use_junction=args.junction, Tm_opt=args.Tm_opt)
        else:
            sites = make_sites.build_by_seq(query=query_string, primer_type=args.type, \
                primer_num_return=primer_num_return, size_min=args.product_size_min, size_max=args.product_size_max, \
                    pick_internal=args.pick_oligo, Tm_opt=args.Tm_opt)
        if 'error' in sites:
            raise Exception(sites['error'])
        primers = design_primer.multiple(sites, cpu=args.cpu)

    ###################  Checking specificity  #############
    if args.run_mode!='design':
        primers = run_blast.run_blast_parallel(primers=primers, dbs=dbs, cpu=args.cpu,\
            checking_size_max=args.checking_size_max, checking_size_min=args.checking_size_min, \
                report_amplicon_seq=args.report_amplicon_seqs, Tm_diff=args.Tm_diff, use_3_end=args.use_3_end, \
                    max_amplicon=args.amplicon_num_max)
        primers = sort_primers.sort_rank(primers=primers, dbs=dbs, max_num_return=args.primer_num_retain, use_isoforms=args.isoform)

    ###################  Checking multiplex  ###############
    dimers = {}
    if args.check_multiplex is True:
        dimers = multiplex.extract_fake_pair(primers, Tm_diff=args.Tm_diff, cpu=args.cpu)

    ###################  Output  ###########################
    region_type = args.type if 'type' in args else 'NA'
    if args.out is None:
        output_fh = open(f'{args.query.name}.json', mode='w')
    else:
        output_fh = args.out
    print(json.dumps({'meta':{'mode':args.run_mode, 'dbs':dbs, 'region_type': region_type, 'check_multiplex':args.check_multiplex}, \
        'primers':primers, 'dimers':dimers}, indent=4), file=output_fh)

    if args.tsv is None:
        tsv_fh = open(f'{args.query.name}.tsv', mode='w')
    else:
        tsv_fh = args.tsv
    print(output.tsv(primers, dbs), file=tsv_fh)
    if args.check_multiplex is True:
        print(output.dimer_list(dimers), file=tsv_fh)


def main():
    global_var.init()
    args = make_args()
    check_environments(args)
    check_templates(args)
    check_qPCR(args)
    run(args)

if __name__ == "__main__":
    main()




import os
import json
import time

from flask import Blueprint, request, current_app, Response

from primerserver2.web.config import load
from primerserver2.core import make_sites, make_primers, design_primer, run_blast, sort_primers, output, global_var, multiplex

# for running progress
global_var.init()
def progress():
    while True:
        yield "data:" + json.dumps({'complete_count': global_var.complete_count, \
            'all_tasks_num': global_var.all_tasks_num, 'current_task': global_var.current_task}) + '\n\n'
        time.sleep(1)
        # if global_var.complete_count>0 and global_var.complete_count==global_var.all_tasks_num:
        #     break
        if global_var.current_task=='finish':
            break
    yield "data:" + json.dumps({'complete_count': global_var.complete_count, \
        'all_tasks_num': global_var.all_tasks_num, 'current_task': global_var.current_task}) + '\n\n'

bp = Blueprint('run', __name__)
@bp.route('/run', methods=['POST'])
def run():
    
    ###################  init #############################
    web_config = load()
    global_var.init()
    db_dir = web_config['templates_directory']

    ###################  Design primers ###################
    query_string = request.form['query']
    dbs = [db_dir+'/'+x for x in request.form['selected_dbs'].split(',')]
    if request.form['app-type']=='check':
        primers = make_primers.make_primers(query=query_string)
        if 'error' in primers:
            return json.dumps({'error': '<p>'+primers['error']+'</p>'+'<p>Your inputs might be invalid; Check the \
                <a href="javascript:void(0)" data-toggle="modal" data-target="#input-help">manual</a> carefully \
                to ensure that your inputs are in correct formats</p>'}, indent=4)
    else:
        if request.form['app-type']=='design':
            primer_num_return = request.form['retain']
        else:
            primer_num_return = request.form['primer_num_return']
        if make_sites.judge_input_type(query_string)=='pos':
            sites = make_sites.build_by_pos(query=query_string, template_file=dbs[0], primer_type=request.form['region_type'], \
                primer_num_return=int(primer_num_return), size_min=int(request.form['product_size_min']), \
                    size_max=int(request.form['product_size_max']))
        else:
            sites = make_sites.build_by_seq(query=query_string, primer_type=request.form['region_type'], \
                primer_num_return=int(primer_num_return), size_min=int(request.form['product_size_min']), \
                    size_max=int(request.form['product_size_max']))
        if 'error' in sites:
            return json.dumps({'error': '<p>'+sites['error']+'</p>'+'<p>Your inputs might be invalid; Check the \
                <a href="javascript:void(0)" data-toggle="modal" data-target="#input-help">manual</a> carefully \
                to ensure that your inputs are in correct formats</p>'}, indent=4)
        primers = design_primer.multiple(sites, cpu=web_config['cpu'], monitor=False)

    ###################  Checking specificity  #############
    if request.form['app-type']!='design':
        primers = run_blast.run_blast_parallel(primers=primers, dbs=dbs, cpu=web_config['cpu'],\
            checking_size_max=int(request.form['checking_size_max']), checking_size_min=int(request.form['checking_size_min']), \
                report_amplicon_seq=bool(int(request.form['report_amplicon_seqs'])), Tm_diff=int(request.form['Tm_diff']), \
                    use_3_end=bool(int(request.form['use_3_end'])), monitor=False)
        primers = sort_primers.sort_rank(primers=primers, dbs=dbs, max_num_return=int(request.form['retain']))

    ###################  Checking multiplex  ###############
    dimers = {}
    if int(request.form['multiplex'])==1:
        dimers = multiplex.extract_fake_pair(primers, Tm_diff=int(request.form['Tm_diff']), cpu=web_config['cpu'], monitor=False)

    ###################  Output    #########################
    global_var.current_task = 'finish'
    return json.dumps({'meta':{'mode':request.form['app-type'], 'dbs':dbs, 'region_type': request.form['region_type'], \
         'check_multiplex': bool(int(request.form['multiplex']))}, 'primers':primers, 'dimers':dimers}, indent=4)
    

@bp.route('/monitor')
def monitor():
    time.sleep(1)
    return Response(progress(), mimetype= 'text/event-stream')

@bp.route('/stop')
def stop():
    global_var.stop_run = True
    return ""


import json
import multiprocessing as mp
import time

from itertools import product
from collections import defaultdict

import primer3
import progressbar

from primerserver2.core import global_var

def judge_two_site(site_pair):

    if global_var.stop_run is True:
        return []

    dimers = []
    for (rank_1, primers_1) in site_pair['site_1'].items():
        for (rank_2, primers_2) in site_pair['site_2'].items():
            for (seq_1, seq_2) in product([primers_1['seq_L'], primers_1['seq_R']], [primers_2['seq_L'], primers_2['seq_R']]):
                min_Tm = min(primers_1['min_Tm'], primers_2['min_Tm'])-site_pair['Tm_diff']

                Heterodimer = primer3.calcHeterodimer(seq_1, seq_2, output_structure=True)
                if Heterodimer.tm > min_Tm:
                    dimers.append([site_pair['id_1'], site_pair['id_2'], rank_1, rank_2, seq_1, seq_2, Heterodimer.tm, Heterodimer.ascii_structure])
                    break
                EndStability = primer3.bindings.calcEndStability(seq_1, seq_2)
                if EndStability.tm > min_Tm:
                    dimers.append([site_pair['id_1'], site_pair['id_2'], rank_1, rank_2, seq_1, seq_2, EndStability.tm, ''])
                    break
    return dimers


def extract_fake_pair(primers_dict, Tm_diff=10, cpu=2, monitor=True):

    if global_var.stop_run is True:
        return {'error': 'Stop running'}

    # make sure more than two sites has primers
    site_num = 0
    for (id, primers) in primers_dict.items():
        if primers['PRIMER_PAIR_NUM_RETURNED']>0:
            site_num += 1
    if site_num < 2:
        return {'error': f'Only {site_num} site has primers. Skip the multiplex module'}
    
    # data group
    primer_sites = defaultdict(dict)
    for (id, primers) in primers_dict.items():
        if primers['PRIMER_PAIR_NUM_RETURNED']==0:
            continue
        if 'PRIMER_PAIR_NUM_RETURNED_FINAL' in primers:
            primer_num = primers['PRIMER_PAIR_NUM_RETURNED_FINAL']
        else:
            primer_num = primers['PRIMER_PAIR_NUM_RETURNED']
        for amplicon_rank in range(0, primer_num):
            if 'PRIMER_PAIR_NUM_RETURNED_FINAL' in primers:
                raw_rank = primers[f'PRIMER_PAIR_AMPLICON_NUM_RANK_{amplicon_rank}']
            else:
                raw_rank = amplicon_rank
            seq_L = primers[f'PRIMER_LEFT_{raw_rank}_SEQUENCE']
            seq_R = primers[f'PRIMER_RIGHT_{raw_rank}_SEQUENCE']
            min_Tm = min(primers[f'PRIMER_LEFT_{raw_rank}_TM'], primers[f'PRIMER_RIGHT_{raw_rank}_TM'])
            primer_sites[id][amplicon_rank] = {"seq_L": seq_L, "seq_R": seq_R, "min_Tm": min_Tm}

    # distribute tasks
    pool = mp.Pool(processes=cpu)
    multi_res = []
    global_var.all_tasks_num = 0
    global_var.current_task = 'multiplex'
    primer_ids = list(primer_sites.keys())
    for i in range(0, len(primer_ids)):
        id_1 = primer_ids[i]
        for j in range(i+1, len(primer_ids)):
            id_2 = primer_ids[j]
            site_pair = {'id_1':id_1, 'site_1':primer_sites[id_1], 'id_2':id_2, 'site_2':primer_sites[id_2], 'Tm_diff':Tm_diff}
            global_var.all_tasks_num += 1
            multi_res.append(pool.apply_async(judge_two_site, (site_pair,)))

    # monitor
    if monitor is True:
        widgets = ['Checking Multiplex: ', progressbar.Counter(),\
            ' Finished', ' (', progressbar.Percentage(), ')', \
                progressbar.Bar(), progressbar.ETA()]
        bar = progressbar.ProgressBar(widgets=widgets, max_value=global_var.all_tasks_num).start()

    while True:
        global_var.complete_count = sum([1 for x in multi_res if x.ready()])
        if global_var.complete_count == global_var.all_tasks_num:
            if monitor is True:
                bar.finish()
            break
        if monitor is True:
            bar.update(global_var.complete_count)
        time.sleep(0.1)

    # results
    dimers_dict = defaultdict(dict)
    for result in multi_res:
        for dimer in result.get():
            (id_1, id_2, rank_1, rank_2, seq_1, seq_2, tm, ascii_structure) = dimer
            dimers_dict[id_1][rank_1] = defaultdict(dict)
            dimers_dict[id_1][rank_1][id_2] = defaultdict(dict)
            dimers_dict[id_1][rank_1][id_2][rank_2] = {'seq_1':seq_1, 'seq_2':seq_2, 'Tm':round(tm, 2), 'ascii_structure':ascii_structure}
    return dimers_dict

if __name__ == "__main__":
    global_var.init()
    primers_dict = json.load(open('tests/_internal_/multisites.json'))
    dimers_dict = extract_fake_pair(primers_dict['primers'])
    print(json.dumps(dimers_dict, indent=4))


#!/usr/bin/env python3

'''PrimerServer2: a high-throughput primer design and specificity-checking platform
Github: https://github.com/billzt/PrimerServer2
This is a CLI script to configure PrimerServer2
'''

import argparse
import os
import sys
import json

def load():
    home_dir = os.environ['HOME']
    web_config = json.load(open(f'{home_dir}/.primerserver.json'))
    return web_config

def prepare():
    home_dir = os.environ['HOME']
    if os.path.isfile(f'{home_dir}/.primerserver.json') is False:
        data_path = os.path.join(os.path.dirname(__file__), '../data')
        os.system(f"cp {data_path}/web_config_sample.json {home_dir}/.primerserver.json")
    print(f'Configure file {home_dir}/.primerserver.json is ready. Please modify it')

def check():
    home_dir = os.environ['HOME']
    status = 'error'
    msg = ''
    if os.path.isfile(f'{home_dir}/.primerserver.json') is False:
        msg = f'The configure file {home_dir}/.primerserver.json is not found. Perhaps you are in a fresh \
            installation? Please run "primerserver-config" to load and edit your configure files'
        return {'status': status, 'msg': msg}
    web_config = json.load(open(f'{home_dir}/.primerserver.json'))
    for key in ['cpu', 'templates_directory', 'templates']:
        if key not in web_config:
            msg = f'The configure file {home_dir}/.primerserver.json is broken. It lacks the "{key}" key.'
            return {'status': status, 'msg': msg}
    if isinstance(web_config['cpu'], int) is False:
        msg = f'There is some error in the configure file {home_dir}/.primerserver.json: the value for cpu must \
            be int.'
        return {'status': status, 'msg': msg}
    db_dir = web_config['templates_directory']
    if os.path.isdir(db_dir) is False:
        msg = f'There is some error in the configure file {home_dir}/.primerserver.json: the path for \
            templates_directory ({db_dir}) does not exist'
        return {'status': status, 'msg': msg}
    for (dbname, dbinfo) in web_config['templates'].items():
        if os.path.isfile(f'{db_dir}/{dbname}') is False:
            if dbname=='example.fa' or dbname=='example2.fa':
                data_path = os.path.join(os.path.dirname(__file__), '../data')
                os.system(f"cp {data_path}/{dbname}* {db_dir}/")
            else:
                msg = f'File {db_dir}/{dbname} does not exist'
                return {'status': status, 'msg': msg}
        if os.path.isfile(f'{db_dir}/{dbname}.nhr') is False:
            msg = f'File {db_dir}/{dbname}.nhr does not exist. Use "makeblastdb" to index {db_dir}/{dbname}'
            return {'status': status, 'msg': msg}
        if os.path.isfile(f'{db_dir}/{dbname}.fai') is False:
            msg = f'File {db_dir}/{dbname}.fai does not exist. Use "samtools faidx" to index {db_dir}/{dbname}'
            return {'status': status, 'msg': msg}
        for key in ['IDs', 'description', 'group']:
            if key not in dbinfo:
                msg = f'The configure file {home_dir}/.primerserver.json is broken. It lacks the "{key}" key.'
                return {'status': status, 'msg': msg}
            if isinstance(dbinfo[key], str) is False:
                msg = f'There is some error in the configure file {home_dir}/.primerserver.json: the value for {key} must \
                    be strings'
                return {'status': status, 'msg': msg}
    status = 'success'
    return {'status': status, 'msg': msg}
    
if __name__ == "__main__":
    print(json.dumps(check()))



import json
import re
import sys

from operator import itemgetter

import progressbar

bar = progressbar.ProgressBar(max_value=progressbar.UnknownLength)

def get_junction(gff_file, features=['exon'], rna_print_ID_key='Parent', remove_version=True, monitor=True):
    with open(gff_file) as f:
        id2exons = {}
        id2strand = {}
        id2print = {}
        i = 0
        for line in f:
            if line.startswith('#'):
                continue
            (feature, start, end, strand, note) = itemgetter(2,3,4,6,-1)(line.strip().split('\t'))
            if feature not in features:
                continue
            notes = note.split(';')
            for x in notes:
                if 'Parent=' in x:
                    rna_ID = x.replace('Parent=', '')
                    if rna_ID not in id2exons:
                        id2exons[rna_ID] = []
                    id2exons[rna_ID].append([int(start), int(end)])
                    id2strand[rna_ID] = strand
                if f'{rna_print_ID_key}=' in x:
                    rna_print_ID = x.replace(f'{rna_print_ID_key}=', '')
                    if remove_version is True:
                        rna_print_ID = re.sub(r'\.\d+$', '', rna_print_ID)
                    id2print[rna_ID] = rna_print_ID
            if monitor is True:
                bar.update(i)
            i+=1
        
        junction_data = {}
        i = 0
        for rna_ID in id2print.keys():
            exons = id2exons[rna_ID]
            if len(exons)==1:
                continue
            strand = id2strand[rna_ID]
            rna_print_ID = id2print[rna_ID]
            
            # sort exons according to strand
            if strand=='-':
                exons = sorted(exons, key=lambda x:x[0], reverse=True)
            
            # junctions
            junction_poses = []
            junction_pos = 0
            for i in range(0, len(exons)-1):
                (start1, end1) = exons[i]
                (start2, end2) = exons[i+1]
                intron_length = start2-end1-1 if strand=='+' else start1-end2-1
                junction_pos += end1-start1+1
                junction_poses.append({'pos': junction_pos, 'intron_len': intron_length})
            junction_data[rna_print_ID] = {'strand': strand, 'junctions':junction_poses}
            if monitor is True:
                bar.update(i)
            i+=1

        return junction_data
            

def get_AS(gff_file, features=['mRNA'], rna_print_ID_key='ID', remove_version=True, monitor=True):
    with open(gff_file) as f:
        gene2rnas = {}
        rna2gene = {}
        i = 0
        for line in f:
            if line.startswith('#'):
                continue
            (feature, note) = itemgetter(2,-1)(line.strip().split('\t'))
            if feature not in features:
                continue
            notes = note.split(';')
            for x in notes:
                if 'Parent=' in x:
                    gene_ID = x.replace('Parent=', '')
                    if gene_ID not in gene2rnas:
                        gene2rnas[gene_ID] = []
                if f'{rna_print_ID_key}=' in x:
                    rna_print_ID = x.replace(f'{rna_print_ID_key}=', '')
                    if remove_version is True:
                        rna_print_ID = re.sub(r'\.\d+$', '', rna_print_ID)
            rna2gene[rna_print_ID] = gene_ID
            gene2rnas[gene_ID].append(rna_print_ID)
            if monitor is True:
                bar.update(i)
            i+=1
        
        rna2partners = {}
        i = 0
        for (rna_print_ID, gene_ID) in rna2gene.items():
            partners = gene2rnas[gene_ID]
            if len(partners)>1:
                rna2partners[rna_print_ID] = partners
            if monitor is True:
                bar.update(i)
            i+=1
        return rna2partners



if __name__ == "__main__":
    junction_data = get_junction(gff_file='tests/_internal_/ref_GRCh38.p12_top_level.gff3', rna_print_ID_key='transcript_id')
    as_data = get_AS(gff_file='tests/_internal_/ref_GRCh38.p12_top_level.gff3', rna_print_ID_key='transcript_id')
    #print(json.dumps(junction_data, indent=4))
    print(json.dumps(as_data, indent=4))


'''PrimerServer2: a high-throughput primer design and specificity-checking platform

Github: https://github.com/billzt/PrimerServer2

PrimerServer was proposed to design genome(or transcriptome)-wide specific PCR primers. It uses candidate primers 
produced by Primer3, uses BLAST and nucleotide thermodynamics to search for possible amplicons and filters out 
specific primers for each site. By using multiple CPUs, it runs very fast, ~0.4s per site in our case study for 
more than 10000 sites.

External Dependencies:

blastn and makeblastdb (from NCBI BLAST+)

samtools (>=v1.9)

'''



from setuptools import setup, find_packages

from primerserver2.core import version

setup(name='primerserver2',
      version=version.get(),
      description="a high-throughput primer design and specificity-checking platform",
      long_description=__doc__,
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Operating System :: Unix',
          'Programming Language :: Python :: 3.6',
          'Topic :: Scientific/Engineering :: Bio-Informatics'
      ],
      keywords='primer bioinformatics PCR',
      author='Tao Zhu',
      author_email='taozhu@mail.bnu.edu.cn',
      url='https://github.com/billzt/PrimerServer2',
      license='MIT',
      packages=find_packages(),
      include_package_data=True,
      python_requires='>=3.6',
      install_requires=[
          'primer3-py',
          'progressbar2',
          'flask',
          'python-dotenv'
      ],
      entry_points={
          'console_scripts': [
              'primertool = primerserver2.cmd.primertool:main',
              'primertool-junction = primerserver2.cmd.junctions:main',
              'primertool-isoform = primerserver2.cmd.isoforms:main',
              'primerserver-config = primerserver2.web.config:prepare'
          ]
      },
)



import re
import json

import primer3

from primerserver2.core import global_var
from primerserver2.core.Santalucia_NN_Tm import transform_degenerate

def calculate_GC(seq):
    gc = len([x for x in seq.upper() if x=='G' or x=='C'])
    return gc/len(seq)*100

def check_primer_seq(seq):
    if len(seq)<10 or len(seq)>40:
        return False
    if re.search('[^ATGCNRYMKSWHBVD]', seq, re.RegexFlag.IGNORECASE) is not None:
        return False

def make_primers(query):
    '''
        In case of user's input, make list of primer dicts similar as the result of design_primer module
        Input:
            query: a string in multi-lines
    '''
    if global_var.stop_run is True:
        return {'error': 'Stop running'}

    primers = {}

    for line in query.splitlines():
        line_data = re.split(r'\s+', line.strip())
        if len(line_data)==4:
            (id, rank, seq_F, seq_R) = line_data
            rank = int(rank)
        elif len(line_data)==3:
            (id, seq_F, seq_R) = line_data
            rank = 0
        else:
            return {'error': f'Your input: {line} does not have three or four columns'}
        
        # whether seq_F and seq_R are valid primer seqs
        if check_primer_seq(seq_F) is False:
            return {'error': f'Your input: {seq_F} does not seem like a vaild primer seq'}
        if check_primer_seq(seq_R) is False:
            return {'error': f'Your input: {seq_R} does not seem like a vaild primer seq'}
        
        # generate primers
        if id not in primers:
            primers[id] = {}
        primers[id]['PRIMER_PAIR_NUM_RETURNED'] = rank+1
        primers[id][f'PRIMER_PAIR_{rank}_PENALTY'] = 0
        primers[id][f'PRIMER_LEFT_{rank}_SEQUENCE'] = seq_F
        primers[id][f'PRIMER_RIGHT_{rank}_SEQUENCE'] =  seq_R
        primers[id][f'PRIMER_LEFT_{rank}'] = [-1, len(seq_F)]
        primers[id][f'PRIMER_RIGHT_{rank}'] = [-1, len(seq_R)]
        primers[id][f'PRIMER_LEFT_{rank}_TM'] = primer3.calcTm(transform_degenerate(seq_F))
        primers[id][f'PRIMER_RIGHT_{rank}_TM'] = primer3.calcTm(transform_degenerate(seq_R))
        primers[id][f'PRIMER_LEFT_{rank}_GC_PERCENT'] = calculate_GC(transform_degenerate(seq_F))
        primers[id][f'PRIMER_RIGHT_{rank}_GC_PERCENT'] = calculate_GC(transform_degenerate(seq_R))
        primers[id][f'PRIMER_PAIR_{rank}_PRODUCT_SIZE'] = -1
        primers[id]['SEQUENCE_RELATIVE_TARGET_START'] = 0
    
    return primers

if __name__ == "__main__":
    global_var.init()
    with open('tests/_internal_/query_check') as f:
        primers = make_primers(f.read())
        print(json.dumps(primers, indent=4))




import json
import os
import time
import multiprocessing as mp
import sys

import primer3
import progressbar

from primerserver2.core import global_var

def single(site):
    '''
        Using the primer3-module to design primers for a single site, return a list (one or two items)
        in DICT format

        Site format: 
        site['id'], site['template'], site['type'], site['pos'], site['length'], 
        site['size_max'], site['size_min'], site['primer_num_return'], site['pick_internal']
    '''
    if global_var.stop_run is True:
        return []

    id = site['id']
    template = site['template']
    type = site['type']
    pos = site['pos']
    length = site['length']
    size_max = site['size_max']
    size_min = site['size_min']
    primer_num_return = site['primer_num_return']
    Tm_opt = site['Tm_opt'] if 'Tm_opt' in site else 60

    p3_json = os.path.join(os.path.dirname(__file__), '../data/p3_settings.json')
    p3_settings = dict(json.load(open(p3_json)))
    p3_settings['PRIMER_PRODUCT_SIZE_RANGE'] = [[size_min, size_max]]
    p3_settings['PRIMER_NUM_RETURN'] = primer_num_return
    p3_settings['PRIMER_OPT_TM'] = Tm_opt
    p3_settings['PRIMER_MIN_TM'] = Tm_opt-3
    p3_settings['PRIMER_MAX_TM'] = Tm_opt+3
    p3_settings['PRIMER_INTERNAL_OPT_TM'] = Tm_opt+10
    p3_settings['PRIMER_INTERNAL_MIN_TM'] = p3_settings['PRIMER_INTERNAL_OPT_TM']-3
    p3_settings['PRIMER_INTERNAL_MAX_TM'] = p3_settings['PRIMER_INTERNAL_OPT_TM']+3
    p3_settings['PRIMER_MAX_SELF_ANY_TH'] = Tm_opt-15
    p3_settings['PRIMER_PAIR_MAX_COMPL_ANY_TH'] = Tm_opt-15
    p3_settings['PRIMER_MAX_SELF_END_TH'] = Tm_opt-15
    p3_settings['PRIMER_PAIR_MAX_COMPL_END_TH'] = Tm_opt-15
    p3_settings['PRIMER_MAX_HAIRPIN_TH'] = Tm_opt-15
    if site['pick_internal'] is True:
        p3_settings['PRIMER_PICK_INTERNAL_OLIGO'] = 1

    #  mode: target or include
    if type=='SEQUENCE_TARGET' or type=='SEQUENCE_INCLUDED_REGION':
        primer3.bindings.setP3Globals(p3_settings)
        if len(site['junction'])==0:
            p3 = primer3.bindings.designPrimers(
                {
                    'SEQUENCE_ID': id,
                    'SEQUENCE_TEMPLATE': template,
                    type: [pos,length]
                })
            p3['SEQUENCE_RELATIVE_TARGET_START'] = pos
            return [{'id': id, 'primers': p3}]
        else:
            # qPCR: primer pairs separated by at least one intron on the corresponding genomic DNA
            p3_A = primer3.bindings.designPrimers(
                {
                    'SEQUENCE_ID': id+'-PARTA',
                    'SEQUENCE_TEMPLATE': template,
                    'SEQUENCE_INCLUDED_REGION': [pos,length],
                    'SEQUENCE_TARGET': [[x,1] for x in site['junction']]
                })
            p3_A['SEQUENCE_RELATIVE_TARGET_START'] = pos

            # qPCR: primers must span an exon-exon junction
            p3_B = primer3.bindings.designPrimers(
                {
                    'SEQUENCE_ID': id+'-PARTB',
                    'SEQUENCE_TEMPLATE': template,
                    'SEQUENCE_INCLUDED_REGION': [pos,length],
                    'SEQUENCE_OVERLAP_JUNCTION_LIST': site['junction']
                })
            p3_B['SEQUENCE_RELATIVE_TARGET_START'] = pos
        
            return([{'id': id+'-PARTA', 'primers': p3_A},
                    {'id': id+'-PARTB', 'primers': p3_B}
            ])

    #  mode: force
    if type=='FORCE_END':
        p3_settings['PRIMER_MIN_LEFT_THREE_PRIME_DISTANCE'] = -1
        p3_settings['PRIMER_MIN_RIGHT_THREE_PRIME_DISTANCE'] = 3
        primer3.bindings.setP3Globals(p3_settings)
        p3_L = primer3.bindings.designPrimers(
            {
                'SEQUENCE_ID': id+'-PARTA',
                'SEQUENCE_TEMPLATE': template,
                'SEQUENCE_FORCE_LEFT_END': pos
            })
        p3_L['SEQUENCE_RELATIVE_TARGET_START'] = pos

        p3_settings['PRIMER_MIN_LEFT_THREE_PRIME_DISTANCE'] = 3
        p3_settings['PRIMER_MIN_RIGHT_THREE_PRIME_DISTANCE'] = -1
        primer3.bindings.setP3Globals(p3_settings)
        p3_R = primer3.bindings.designPrimers(
            {
                'SEQUENCE_ID': id+'-PARTB',
                'SEQUENCE_TEMPLATE': template,
                'SEQUENCE_FORCE_RIGHT_END': pos
            })
        p3_R['SEQUENCE_RELATIVE_TARGET_START'] = pos

        return([{'id': id+'-PARTA', 'primers': p3_L},
                {'id': id+'-PARTB', 'primers': p3_R}
        ])


def multiple(sites, cpu=2, monitor=True):
    if global_var.stop_run is True:
        return {'error': 'Stop running'}
    
    # distribute
    pool = mp.Pool(processes=cpu)
    multi_res = []
    global_var.all_tasks_num = 0
    global_var.current_task = 'design'
    for site in sites:
        global_var.all_tasks_num += 1
        multi_res.append(pool.apply_async(single, (site,)))

    # monitor
    if monitor is True:
        widgets = ['Designing Primers: ', progressbar.Counter(), ' Finished', ' (', progressbar.Percentage(), ')', \
            progressbar.Bar(), progressbar.ETA()]
        bar = progressbar.ProgressBar(widgets=widgets, max_value=global_var.all_tasks_num).start()

        while True:
            global_var.complete_count = sum([1 for x in multi_res if x.ready()])
            if global_var.complete_count == global_var.all_tasks_num:
                bar.finish()
                break
            bar.update(global_var.complete_count)
            time.sleep(0.1)

    # results
    primers = {}
    for result in multi_res:
        primers_in_each_site = result.get()
        for primer in primers_in_each_site:
            primers[primer['id']] = primer['primers']
        
    return primers



if __name__ == "__main__":
    global_var.init()
    template = ''
    with open('tests/example.fa') as f:
        data = f.read().split('>')
        seq1 = data[1].splitlines()
        del seq1[0]
        template = ''.join([x.strip() for x in seq1])
        print(json.dumps(multiple([
                {
                    'id':'A-SEQUENCE_INCLUDED_REGION-1-600', 
                    'template':template,
                    'type':'SEQUENCE_INCLUDED_REGION',
                    'pos':1,
                    'length':600,
                    'size_min':75,
                    'size_max':1000,
                    'primer_num_return':10,
                    'pick_internal': False,
                    'junction': []
                }, 
                {
                    'id':'B-SEQUENCE_TARGET-200-10', 
                    'template':template,
                    'type':'SEQUENCE_TARGET',
                    'pos':200,
                    'length':10,
                    'size_min':75,
                    'size_max':1000,
                    'primer_num_return':10,
                    'pick_internal': False,
                    'junction': []
                },
                {
                    'id':'C-FORCE_END-319-1', 
                    'template':template,
                    'type':'FORCE_END',
                    'pos':319,
                    'length':1,
                    'size_min':75,
                    'size_max':1000,
                    'primer_num_return':10,
                    'pick_internal': False,
                    'junction': []
                }
                ]), indent=4))
